(*
  QS

  Luc Maranget, projet Moscova, Inria Rocquencourt
 
 Copyright 2005 Institut National de Recherche en Informatique et    
 en Automatique.  All rights reserved.  This file is distributed     
 under the terms of the Q Public License version 1.0.          
*)

{
open Printf
open Lexing

let verbose = ref (Array.length Sys.argv > 2)

module Ints =
  Set.Make
    (struct
      type t = int
      let compare = Pervasives.compare
    end)
;;

let add_primes b a ps r =
  let small, big = List.partition (fun p -> p < b) ps in
  (a, small, big)::r

let nlines = ref 1
}  

let xpat =  ('-' ? ['0'-'9']+ as a) ('[' [^']']* ']') ? ':'
let ipat = ['0'-'9']+

rule main = parse
| (ipat as n) ' '+ (ipat as bs) '\n'
  {let b = int_of_string bs in
  incr nlines ;
   let m = main_cont b [] lexbuf in
   n,bs,m }
| "" {failwith "main"} 

and main_cont b r = parse 
| eof { r }
| '*'? xpat
    {let ps = primes [] lexbuf in
    main_cont b (add_primes b a ps r) lexbuf}
| "" {skip lexbuf ; main_cont b r lexbuf}

and primes r = parse
| ' '+ {primes r lexbuf}
| "-1"|['0'-'9']+ as lxm
    {let p = int_of_string lxm in
    primes (p::r) lexbuf}
| '\n' { incr nlines ; r }
| "" { failwith ("primes: "^string_of_int !nlines) }

and skip = parse
| [^'\n']* '\n' { incr nlines }
| "" { failwith "skip" }

{
let name = if Array.length Sys.argv > 1 then Sys.argv.(1) else ""
;;

let hashtbl_card t =  Hashtbl.fold (fun _ _ r -> r+1) t 0

let weight xs =
  let rec do_rec p odd = function
    | [] -> if odd then 1 else 0
    | q::rem ->
        if p=q then
          do_rec p (not odd) rem
        else if odd then 1+do_rec q true rem
        else do_rec q true rem in
  match xs with
  | [] -> 0
  | p::rem -> do_rec p true rem

let rec restore n xs =
  let rec do_rec r k xs =
    if k = 0 then List.rev r
    else match xs with
    | (x,_)::rem -> do_rec (x::r) (k-1) rem
    | _ -> assert false in
  do_rec [] n xs
      

let reduce2 np primes =
  let primes = List.rev_map (fun ((_,ps,_) as c) -> c,weight ps) primes in
  let primes = List.sort (fun (_,c1) (_,c2) -> c1-c2) primes in
  restore (np+10) primes

let rec reduce1 primes =
  let t = Hashtbl.create 101 in
  let add_prime p =
  try
    let r = Hashtbl.find t p in
    r := !r+1
  with
  | Not_found ->
      Hashtbl.add t p (ref 1) in

  List.iter (fun (_,ps,_) -> List.iter add_prime ps) primes ;
  let np = hashtbl_card t in
  let single =
    Hashtbl.fold (fun p c r ->
      match !c with
      | 1 -> Ints.add p r
      | _ -> r) t Ints.empty in
  if Ints.is_empty single then begin
    let nr = List.length primes in
    if nr <= np + 10 then
      primes
    else
      reduce1 (reduce2 np primes)
  end else begin
    let new_primes =
      List.filter
         (fun ((_,ps,_)) ->
           List.for_all (fun p -> not (Ints.mem p single)) ps)
        primes in
    reduce1 new_primes
  end
;;

  
let zyva name =
    let chan = match name with "" -> stdin | _ -> open_in name in
  let n,b,primes = main (Lexing.from_channel chan) in
  let primes = reduce1 primes in
  printf "%s %s\n" n b ;
  List.iter
    (fun (a,ps,qs) ->
      begin match qs with
      | [] -> ()
      | _  -> printf "*"
      end ;
      printf "%s:" a ;
      List.iter (fun p -> printf " %i" p) (List.rev ps) ;
      List.iter (fun p -> printf " %i" p) qs ;
      printf "\n")
    primes ;
  ()
;;

let _ = zyva name ; exit 0
;;


  
} 
